/*

   d-exp - Interfacing to a DECtalk via C

   This is derived from, and should be functionally equivalent to,
   dtk-exp in Emacspeak 6.0 by T. V. Raman.  It will also drive
   previous versions of Emacspeak.

   $Id: d-exp.c,v 1.9 1997/08/06 01:55:01 jrv Exp jrv $

   $Log: d-exp.c,v $
   Revision 1.9  1997/08/06 01:55:01  jrv
    dectalk_get_acknowledgement: waiting until something becomes readable.
    dectalk_gobble_acknowledgements: wait up to 1 ms for more text.
    optionally writing to dfile: time and exact input data string
    log file names changed from d-exp.log to emacspeak.ilog,
    queue_speech: fixed calculation of q_head (though bug never exercised?)
    readline(): fixed calculation of remaining buffer size, replacing
   newline by space rather than deleting, continuing to read if brace
   depth is positive, enlarging buffer only if current buffer is full.
    Switch -V speaks program version number.
    Initial bufsiz increased from 64 to 256 bytes.
    Accept old as well as new style emacspeak commands.

   Revision 1.8.0.10  1997/05/30 01:41:14  jrv
   Simplified parsing.  Input commands need not be followed by a space.

   Revision 1.8.0.9  1997/05/30 01:29:07  jrv
     In dectalk-say and dectalk_letter, immediately after clearing
   `talking?', call queue_clear.
     In dectalk_tone, writing tone command with puts rather than queue_speech.
     In speech_task, no longer writing "interrupt" or "done".
     In main program, setting up so writes to output port are unbuffered
   instead of line buffered.

   Renamed four commands in input stream:
     dectalk_letter -> l
     dectalk_stop -> s
     dectalk_tone -> t
     queue_speech -> q

   Revision 1.8.0.8  1997/05/15 01:47:08  jrv
     Logging of input and output is now controlled by environment
   variable EMACSPEAK rather than compile time switches.
     Saving pointer returned by dectalk_get_acknowledgement, and freeing
   buffer (fixes memory leak).
     -s switch to speak phrase given on command line
     -v switch to print program version to stdout
     Have replacement implementation of strstr.

   Revision 1.8.0.7  1997/01/27 00:57:24  jrv
   Reset string contains d-exp version number.
   Waits are now 10 msec.
   dectalk_set_rate command includes close bracket.
   Fixed memory leak in call to dectalk_get_acknowledgement().

   Revision 1.8.0.6  1997/01/25 02:37:10  jrv
   Optionally copies output to a file in addition to the synthesizer.
   Startup string includes driver name and version.
   Eliminated extra characters in output for setting punctuation mode.
   Added another blank before [:say clause].
   Writing "interrupt" back to parent.

   Revision 1.8.0.5  1997/01/23 02:59:47  jrv
   perror calls include source code line number

   Revision 1.8.0.4  1997/01/22 03:07:14  jrv
   Shrank some local character string buffers.
   Removed one xmalloc call (using automatic variable instead).
   xmalloc exits on error.

   Revision 1.8.0.4  1997/01/22 01:44:54  jrv
   *** empty log message ***

   Revision 1.8.0.3  1997/01/18 03:40:07  jrv
   *** empty log message ***

   Revision 1.8.0.2  1997/01/17 03:19:05  jrv
   nothing significant.

   Revision 1.8.0.1  1997/01/16 02:25:54  jrv
   Adapted for automatic configuration.
   Help message shows logfile name.
   Including replacement code for strstr().

   Revision 1.8  1996/12/19 03:02:04  jrv
   queue grows dynamically as needed.
   queue is circular buffer.
   Renamed debug file pointer: fd -> dfile.
   Implemented -t switch and the DECTALK environment variable to write to
   trace file instead of the DECtalk.
   No longer writing time into trace file.
   port is now a pointer instead of a character array.
   Eliminated extra braces.

   Revision 1.7  1996/12/14 20:05:37  jrv
   Fixed regexp in queue_speech so dtk-toggle-capitalization now works.
   Put functions in alphabetic order.

   Revision 1.6  1996/12/14 03:47:36  jrv
   read_pending_p() replaced by readable().
   readable() no longer has time parameters.
   Writing to log is controlled by symbol LOG.
   First item written to log is current date and time.

   Revision 1.5  1996/07/27 23:39:57  jrv
   Added some casts.
   buf is allocated from the heap, and grows if needed.
   Detecting and reporting more errors.
   Writing input stream to log file.

   Revision 1.4  1996/07/06 17:24:10  jrv
   In dectalk_tone, eliminated a conversion from text to integers and back.

   Revision 1.3  1996/07/06 17:06:51  jrv
   Using new gsub() calling sequence.
   Not defining four unused functions.
   In dectalk_tone, including escape sequence for tone.
   Using file descriptor for stdin rather than FILE *.
   Corrected sense of empty-queue test.

   Revision 1.2  1996/06/27 21:34:02  jrv
   Functions not called from Emacs declared "static".
   xmalloc() added
   dectalk_split_caps and dectalk_capitalize eliminated (implemented in main
   	command loop).
   Added readline() which enlarges its buffer as needed.
   Added command loop.

   Revision 1.1  1996/06/23 13:58:52  jrv
   Initial revision


   This file is not part of GNU Emacs, but the same permissions apply.
  
   GNU Emacs is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.
  
   GNU Emacs is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
  
   You should have received a copy of the GNU General Public License
   along with GNU Emacs; see the file COPYING.  If not, write to
   the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  */

#include <config.h>
#include <assert.h>
#include <fcntl.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>
#include <syslog.h>
#include <errno.h>

#include <regexp.h>

#ifndef max
#define max(a,b) ((a)>(b)?(a):(b))
#endif

int ilog = 0;			/* nonzero if input is to be logged */
int olog = 0;			/* nonzero if output is to be logged */
static struct timeval then={0,0}; /* last time reported to output log */

char *ifilename = "emacspeak.ilog";
char *ofilename = "emacspeak.olog";

char **queue;			/* pointers to text fragments to be
                                   spoken.  Each fragment was
                                   allocated from the heap, and will
                                   be free'd after being spoken. */
int q_head, q_tail, q_size;	/* indexes into queue */
char punctuations[20]="some";
int say_rate;
int split_caps;
int capitalize;
int space_special_chars;
int speech_rate;
double char_factor = 1.2;
int talking;
int fd;				/* file descriptor for reading from or
				   writing to the speech device */

int ofd;			/* file descriptor for output log */
char *Argv0;
static char *buf;		/* input buffer pointer */
static int bufsiz = 0;		/* # bytes in input buffer */

int gsub(char *exp, char *replacement, char **s);

/*
   these are called from Emacs
*/
static void dectalk_letter(char *text);
static void dectalk_reset(void);
static void dectalk_say(char *text);
static void dectalk_set_character_scale(double new_factor);
static void dectalk_set_punctuations(char *mode);
static void dectalk_set_speech_rate(int rate);
static void dectalk_speak(char *text);
static void dectalk_stop(void);
static void dectalk_tone(char *str);
static void queue_speech(char *element);
/*
   these are only used internally
*/
static char * dectalk_get_acknowledgement(void);
static void dectalk_gobble_acknowledgements(void);
static void queue_clear(void); 
static int queue_empty_p(void); 
static char * queue_remove(void);
static int readable (int fd1, int fd2);
static void speech_task(void);
static void writeout(char *s);
static void *xmalloc(unsigned n);
static void *xrealloc(void *p, unsigned n);

#ifdef NEVER
/* these are unused */
static void dectalk_pause(void);
static void dectalk_resume(void);
static void dectalk_space_special_chars(int flag); 
static void dectalk_synchronize(void);
#endif

#define SENTINAL "Q.Q"

char init_string[] = 
"[:timeout 0]\n"
" [:punc some ]\n"
" [:phoneme arpabet speak on ]\n"
" [:tsr off ]\n"
" [:tsr off ]\n"
" [:power interval 60 ]\n"
" [:sync] [:np :ra 200]\n"
"This is the Dectalk Express.  "
#ifdef FOREIGN			/* Raman includes a foreign language
				   greeting */
"[zhax<15> p'arl],"
" [/dh`ow<100,140> ],  [:np] [  zhax<13>  suw<45>\iy<140,100>]."
#endif
"driver d-exp version " SENTINAL "      .\n"; /* leave space for long
						version number */
char reset_string[] = 
"[:timeout 0]\n"
"[:punc some ]\n"
"[:phoneme arpabet speak on ]\n"
"[:tsr off ]\n"
"[:power interval 60 ]\n"
"d-exp interface version "
SENTINAL 
"      Restoring sanity to the Dectalk Express.\n";

char revision_string[]="$Revision: 1.9 $";
char *revision = revision_string;

/* use syslog facility to record an error encountered on program line
   line.  (stderr probably has nobody watching.) */
static void error(int line)
{
  openlog("d-exp", LOG_PID|LOG_PERROR, LOG_LOCAL0);
  syslog(LOG_ERR, "ver %s line %d: %m", revision, line);
  exit(1);
}

char *
dectalk_get_acknowledgement()
{
  char *retval = NULL;
  fd_set rfds, wfds, efds;
  struct timeval tv;
  int numchars = 0, maxchars = 0;

  FD_ZERO(&rfds);
  FD_SET(fd, &rfds);
  FD_SET(0, &rfds);
  FD_ZERO(&wfds);
  FD_ZERO(&efds);
  (void)select(fd+1, &rfds, &wfds, &efds, NULL);
  if (FD_ISSET(0, &rfds))
    talking = 0;		/* parent has more to say */
  else
    {
      while(readable(fd, fd))
	{
	  if (numchars >= maxchars)
	    {
	      maxchars = maxchars*2 + 32;
	      retval = xrealloc(retval, maxchars);
	    }
	  if (read(fd, retval+numchars++, 1) < 0)
	    error(__LINE__);
	  FD_ZERO(&rfds);
	  FD_SET(fd, &rfds);
	  tv.tv_sec = 0;
	  tv.tv_usec = 1000;
	  (void)select(fd+1, &rfds, &wfds, &efds, &tv);
	}
    }

  if (retval && olog)
    {
      static struct timeval now;
      char *s=retval;
      double diff;
      FILE *dfile = fopen(ofilename,"a");

      if (dfile)
	{
	  gettimeofday(&now, NULL);
	  diff = now.tv_sec - 
	    then.tv_sec + (now.tv_usec - then.tv_usec)/1000000.;
	  then = now;
	  fprintf(dfile, " %.24s %9.6f  read: \"", 
	      ctime((time_t *)&now.tv_sec), diff);
	  while (*s)
	    {
	      if (isprint(*s))
		fprintf(dfile, "%c", *s);
	      else
		fprintf(dfile, "\\%03o", *s);
	      s++;
	    }
	  fprintf(dfile, "\"\n");
	  fclose(dfile);
	}
      else
	error(__LINE__);
    }
  return retval;
}
#ifdef PREV
/*  Gobble up any garbage the DECtalk has returned */
void
dectalk_gobble_acknowledgements()
{
  int junk;
  while(readable(fd, fd))
    if (read(fd, &junk, 1) < 0)
      error(__LINE__);
}
#else

/*  Gobble up any garbage the DECtalk has returned */
void
dectalk_gobble_acknowledgements()
{
  int junk;
  fd_set rfds, wfds, efds;
  struct timeval tv;

  if(readable(fd, fd))
    {
      do
	{
	  if (read(fd, &junk, 1) < 0)
	    error(__LINE__);
	  
	  FD_ZERO(&rfds);
	  FD_SET(fd, &rfds);
	  FD_ZERO(&wfds);
	  FD_ZERO(&efds);
	  tv.tv_sec = 0;
	  tv.tv_usec = 1000;
	} while (select(fd+1, &rfds, &wfds, &efds, &tv));
    }
}
#endif

void dectalk_letter(char *text)
{
  char buf[32];
  while(*text == ' ') text++;	/* skip leading spaces */
  if(talking)
    {
      talking=0;
      queue_clear();
    }
  sprintf(buf, "[:rate %d :say letter :sync]", say_rate);
  writeout(buf); 
  writeout(text); 
  sprintf(buf, "[:rate %d :say clause :sync]", speech_rate);
  writeout(buf);
}

void
dectalk_reset()
{
  dectalk_gobble_acknowledgements();
  writeout(reset_string);
}

void dectalk_say(char *text)
{
  char buf[32];

  if(talking)
    {
      talking=0;
      queue_clear();
    }
  dectalk_gobble_acknowledgements();
  writeout("   [:punc all :say word]");
  writeout(text);
  sprintf(buf, "[:punc %.12s :say clause ]", punctuations);
  writeout(buf);
}

void dectalk_set_character_scale(double new_factor)
{
  say_rate = (int)(speech_rate*new_factor);
}

void dectalk_set_punctuations(char *mode)
{
  char buf[32];
  int n;
  n = strspn(mode, "somenoneall");
  if (n > 6) n = 6;
  sprintf(buf, "[:punc %.*s ]", n, mode);
  sprintf(punctuations, "%.*s", n, mode);
  writeout(buf);
}

void dectalk_set_speech_rate(int rate)
{
  char buf[32];
  sprintf(buf, "[:rate %d]", rate);
  writeout(buf);
  speech_rate = rate;
  say_rate = (int)(rate*char_factor);
}

void dectalk_speak(char *text)
{
  queue_speech(text);
  if(!talking)
    {
      writeout("  [:say clause]");
      speech_task();
    }
}

void dectalk_stop()
{
  fd_set rfds, wfds, efds;
  struct timeval tv;
  char junk;

  queue_clear();
  dectalk_gobble_acknowledgements();
  writeout("\003");
  FD_ZERO(&rfds);
  FD_SET(fd, &rfds);
  FD_ZERO(&wfds);
  FD_ZERO(&efds);
  tv.tv_sec = 1;
  tv.tv_usec = 0;

  if(select(fd+1, &rfds, &wfds, &efds, &tv))
    if (read(fd, &junk, 1) < 0)
      error(__LINE__);
  talking = 0;
  dectalk_gobble_acknowledgements();
}

void
dectalk_tone(char *str)
{
  char *buf = xmalloc(32+strlen(str));
  sprintf(buf, "[_ :tone %s]", str);
  writeout(buf);
  free(buf);
}

int queue_empty_p() {return q_head == q_tail;}

void queue_clear() 
{
  while (q_head != q_tail)
    {
      if (q_head == q_size)
	q_head = 0;
      else
	free(queue[q_head++]);
    }
  q_head = q_tail = 0;
}

char *
queue_remove()
{
  if (queue_empty_p())
    return NULL;
  if (q_head == q_size) 
    q_head = 0;
  return queue[q_head++];
}

void
queue_speech(char *element)
{
  char *t = strdup(element);
  if (t == NULL) error(__LINE__);
				/* first protect against DECtalk going
                                   into spell mode by separating
                                   certain characters with spaces */
  if(strstr(punctuations, "all") != NULL)
    gsub("[%&;()$+=/]", " \0 ", &t);
  else
    gsub("([a-zA-Z])([,.!%&;()$+=/])([a-zA-z])", "\\1 \\2 \\3", &t);
  if (capitalize)
				/* precede capitals with a tone */
    gsub("([A-Z])", "[_ :tone 440 10]\\1", &t);
  if (split_caps)
    gsub("([^ -_A-Z])([A-Z][a-zA-Z]* )", "\\1[_<1>]\\2[,] ", &t);
  gsub("([^ -_A-Z])([A-Z])", "\\1[:pause 1]\\2", &t);
  if (strstr(punctuations, "some") != NULL)
    gsub("--", " [_,]", &t);

/*
  if (ilog)
    {
      FILE *dfile = fopen(ifilename,"a");
      fprintf(dfile, "%s(%d):  after substitutions = \"%s\"\n", 
	      __FILE__, __LINE__, t); 
      fclose(dfile);
    }
*/
      
  if ((q_tail + 1 - q_head + q_size)%q_size == 0)
    {				/* queue is full - must enlarge */
      unsigned n = q_size - q_head; /* number of pointers to move */
      unsigned newsize = q_size*2; /* number of pointers in enlarged
				      queue */

/*
      if (ilog)
	{
	  int i;
	  FILE *dfile = fopen(ifilename,"a");
	  fprintf(dfile, "%s(%d): enlarging string queue from %d to %d\n"
		  "before:", __FILE__, __LINE__, q_size, newsize);
	  for (i = 0; i < q_size; i++)
	    fprintf(dfile, "%s%u", i==q_head?" [":(i==q_tail?"] ":" "), 
		    (unsigned)queue[i]);
	  if (q_tail == q_size) fprintf(dfile, "]");
	  fprintf (dfile, "\n");
	  fclose(dfile);
	}
*/

      queue = xrealloc(queue, newsize*sizeof(char *));
      if (q_head > q_tail)
	{
	  memmove(queue + newsize - n, queue + q_head, 
		  n*sizeof(char *));
	  q_head = newsize - n;
	}
      q_size = newsize;

/*
      if (ilog)
	{
	  int i;
	  FILE *dfile = fopen(ifilename,"a");
	  fprintf(dfile, "after (head=%u, tail=%u):", q_head, q_tail);
	  for (i = 0; i < q_size; i++)
	    fprintf(dfile, "%s%u", i==q_head?" [":(i==q_tail?"] ":" "), 
		    (unsigned)queue[i]);
	  if (q_tail == q_size) fprintf(dfile, "]");
	  fprintf (dfile, "\n");
	  fclose(dfile);
	}
*/

    }
  if(q_tail == q_size) q_tail = 0;
  queue[q_tail++] = t;
}

/* return nonzero if either file descriptor fd1 or fd2 could be read
   without blocking. */
int
readable (int fd1, int fd2)
{  
  fd_set rfds, wfds, efds;
  struct timeval tv;

  FD_ZERO(&rfds);
  FD_SET(fd1, &rfds);
  FD_SET(fd2, &rfds);
  FD_ZERO(&wfds);
  FD_ZERO(&efds);
  tv.tv_sec = 0;       	/* do not wait for the files to become readable */
  tv.tv_usec = 0;
  return select(max(fd1, fd2)+1, &rfds, &wfds, &efds, &tv);
}

/* read one line from stdin, and leave it in buf.  Return zero for
   error or EOF.  Enlarge buffer if needed.
*/
int 
readline()
{
  
  char *s, *sprev;
  if (bufsiz == 0)
    buf = xmalloc(bufsiz = 64);
  s = buf;
  while((s = fgets(sprev=s, buf+bufsiz-s, stdin)))
    {
      s = buf + strlen(buf) - 1;
      if (*s == '\n')
	{
	  int count = 0, c;
	  *s = ' ';		/* treat newline as whitespace */
	  for (s = buf; c = *s; s++)
	    {
	      switch(c)
		{
		case '{': count++; break;
		case '}': count--; break;
		default:
		}
	    }
	  if (count == 0)
	    return 1;
	}
      else
	buf = xrealloc(buf, bufsiz *= 2);
      s = buf + strlen(buf);
    }

  if (ilog)
    {
      FILE *dfile = fopen(ifilename,"a");
      fprintf(dfile, "d-exp: input problem: fgets(0x%x, %d, stdin)=0x%x"
	      " feof(stdin)=%d ferror(stdin)=%d"
	      " errno=%d strerror(%d)=\"%s\"\n", 
	      sprev, bufsiz, s,
	      feof(stdin), 
	      ferror(stdin),
	      errno, errno, strerror(errno)); 
      fclose(dfile);
    }
  return 0;		/* EOF or error */
}

void
speech_task()
{
  char *s, buf[32];
  int index;

  index = 1;
  talking = 1;
  dectalk_gobble_acknowledgements();
  while(talking)
    {
      if(queue_empty_p())
	talking = 0;
      else
	{
	  s = queue_remove();
	  writeout(s);
	  free(s);
	  sprintf(buf, "[:index reply %d  ]\013", index);
	  writeout(buf);
	  s = dectalk_get_acknowledgement();
	  free(s);		/* ignore what the DECtalk sends back */
	  index++;
	  if (readable(0, 0))
	    talking = 0;
	}
    }
}

void
writeout(char *s)
{
  if (write(fd, s, strlen(s)) < 0)
    error(__LINE__);

  if (olog)
    {
      static struct timeval now;
      double diff;
      FILE *dfile = fopen(ofilename,"a");

      if (dfile)
	{
	  gettimeofday(&now, NULL);
	  diff = now.tv_sec - 
	    then.tv_sec + (now.tv_usec - then.tv_usec)/1000000.;
	  then = now;
	  fprintf(dfile, " %.24s %9.6f  \"", 
	      ctime((time_t *)&now.tv_sec), diff);
	  while (*s)
	    {
	      if (isprint(*s))
		fprintf(dfile, "%c", *s);
	      else
		fprintf(dfile, "\\%03o", *s);
	      s++;
	    }
	  fprintf(dfile, "\"\n");
	  fclose(dfile);
	}
      else
	{
	  openlog("d-exp", LOG_PID|LOG_PERROR, LOG_LOCAL0);
	  syslog(LOG_ERR, "line %d: %s: %m", ofilename, __LINE__);
	  exit(1);
	}
/*
	{
	  error(__LINE__);
	}
*/
    }
}

void *xmalloc(unsigned n)
{
  void *u = NULL;
  return xrealloc(u, n);
}

void *xrealloc(void *p, unsigned n)
{
  void *u;
  u = realloc(p, n);
  if (u == NULL)
    {
      error(__LINE__);
      exit(1);
    }
  return u;
}


int
main(int argc, char **argv)
{
  int pid, pip[2];
  char *s, *port, c, *statement=0;
  extern char *optarg;
  int val;
  int say_version = 0;		/* nonzero if we are only speaking the
                                   version numbers */
  int testing = 0;		/* nonzero if writes are to actually
				   go into a file instead of the
				   speech device */

  Argv0 = argv[0];

  {
    char *s;
				/* isolate the version number */
    revision += strlen("$Revision: ");
    if ((s = strchr(revision, ' '))) *s = 0;
				/* insert version number into messages */
    if ((s = strstr(init_string, SENTINAL)))
      memcpy(s, revision, strlen(revision));
    if ((s = strstr(reset_string, SENTINAL)))
      memcpy(s, revision, strlen(revision));
  }

  while ((c = getopt(argc, argv, "t:s:vV")) != -1)
    {
      switch(c)
	{
	case 's':
	  statement = optarg;
	  break;
	case 't':
	  testing = 1;
	  ofilename = optarg;
	  break;
	case 'v':
	  printf("d-dt DoubleTalk driver for Emacspeak, version %s\n", 
		 revision);
	  return(0);
	case 'V':
	  say_version = 1;
	  break;
	case '?':
	default:
	  printf("d-exp - Interfacing to a DECtalk via C\n"
		 "usage: d-exp [-t logfile]\n"
		 "(but usually called by emacs directly)\n");
	  exit(1);
	}
    }

  s = getenv("EMACSPEAK");
  if (s && strcmp(s, "ilog") == 0) ilog = 1;
  if (s && strcmp(s, "olog") == 0) olog = 1;
  if (s && strcmp(s, "log") == 0) 
    {
      ilog = olog = 1;
      ofilename = ifilename;
    }

  if (ilog)
    {
      FILE *dfile = fopen(ifilename,"a");
      time_t now;
      time(&now);
      fprintf(dfile, "%s(%d): starting at %s", 
	      __FILE__, __LINE__, ctime(&now)); 
      fclose(dfile);
    }

  if (olog)
    {
      FILE *dfile = fopen(ofilename,"a");
      time_t now;
      time(&now);
      fprintf(dfile, "%s(%d): starting at %s", 
	      __FILE__, __LINE__, ctime(&now)); 
      fclose(dfile);
    }

  buf = xmalloc(bufsiz = 256);
  queue = xmalloc((q_size = 100)*sizeof(char *));

  /* execute "uname" to find out the operating system */

  if (pipe(pip) < 0)
    error(__LINE__);
  if ((pid = fork()) < 0)
    error(__LINE__);	/* pipe error */
  else if (pid > 0)
    {				/* parent */
      if (close(pip[1]))
	error(__LINE__);
      (void)wait(NULL);
    }
  else
    {				/* child */
      if (close(pip[0]))
	error(__LINE__);
      if (pip[1] != 1)
	if (dup2(pip[1], 1) != 1)
	  error(__LINE__);
      if (execlp("uname","uname",NULL))
	error(__LINE__);
    }
  if (read(pip[0], buf, bufsiz) <= 0)
     strcpy(buf, "Linux");

/* set up no buffering on read, but line buffering on write.  But
    these are all unbuffered! */

  if (strstr(buf, "ULTRIX") || strstr(buf, "OSF1"))
    {
      if ((s = getenv("DTK_PORT")) != NULL)
	port = strdup(s);
      else 
	port = strdup("/dev/tty00");
      fd = open(port, O_RDWR);
      if (fd < 0)
	error(__LINE__);
      sprintf(buf, "stty sane 9600 raw  -echo  <  %.32s", port);
      val = system(buf);
      if (val == 127 || val == -1)
	error(__LINE__);
      sprintf(buf, "stty ixon ixoff  <  %.32s", port);
      val = system(buf);
      if (val == 127 || val == -1)
	error(__LINE__);
    }
  else if (strstr(buf, "SunOS"))
    {
      if ((s = getenv("DTK_PORT")) != NULL)
	port = strdup(s);
      else 
	port = strdup("/dev/ttya");
      fd = open(port, O_RDWR);
      if (fd < 0)
	error(__LINE__);
      sprintf(buf, "stty sane 9600 raw  -echo > %.32s", port);
      val = system(buf);
      if (val == 127 || val == -1)
	error(__LINE__);
      sprintf(buf, "stty ixon ixoff  >  %.32s", port);
      val = system(buf);
      if (val == 127 || val == -1)
	error(__LINE__);
    }      
  else
    {
      if ((s = getenv("DTK_PORT")) != NULL)
	port = strdup(s);
      else 
	port = strdup("/dev/ttyS0");
      fd = open(port, O_RDWR);
      if (fd < 0)
	error(__LINE__);
      sprintf(buf, "stty sane 9600 raw  -echo < %.32s", port);
      val = system(buf);
      if (val == 127 || val == -1)
	error(__LINE__);
				/* linux wants the -echo done separately */
      sprintf(buf, "stty -echo < %.32s", port);
      val = system(buf);
      if (val == 127 || val == -1)
	error(__LINE__);
      sprintf(buf, "stty ixon ixoff < %.32s", port);
      val = system(buf);
      if (val == 127 || val == -1)
	error(__LINE__);
    }


  /* initialize the speech parameters */

  split_caps = 1;
  capitalize = 0;
  space_special_chars = 1;
  talking = 0;
  speech_rate = 425;
  char_factor = 1.2;
  say_rate = speech_rate*char_factor;
  q_head = q_tail = 0;
  strcpy(punctuations, "some");

  signal(SIGINT, SIG_IGN);
  dectalk_gobble_acknowledgements();

  if (say_version)
    {
      printf("d-exp DoubleTalk driver for Emacspeak implemented in C, "
	     "version %s\n", revision);
      sprintf(buf, "d-exp DoubleTalk driver for Emacspeak implemented in C, "
	      "version %s\n"
	      "DECtalk firmware [:version speak]", revision);
      dectalk_speak(buf);
      return 0;
    }

  if (statement)
    {
      dectalk_speak(statement);
      return 0;
    }
      
  if (testing)
    {		     /* open regular file INSTEAD of the DECtalk */
      close(fd);
      fd = open(ofilename, O_RDWR|O_CREAT, 0644);
    }

  writeout(init_string);

  /* the main loop */

  while(readline())
    {
      char *s, *t, *cmd;

      if (ilog)
	{
	  FILE *dfile = fopen(ifilename,"a");
	  fprintf(dfile, "%s(%d): queue=[%d,%d), command = ", 
		  __FILE__, __LINE__, q_head, q_tail); 
	  {
	    char *s = buf;
	    fprintf(dfile, "\"");
	    while (*s)
	      {
		if (isprint(*s))
		  fprintf(dfile, "%c", *s);
		else
		  fprintf(dfile, "\\%03o", *s);
		s++;
	      }
	    fprintf(dfile, "\"\n");
	  }
	  fclose(dfile);
	}

				/* eliminate leading whitespace and
                                   surrounding braces */
      s = strrchr(buf, '}');
      if (s != NULL) *s = 0;
      s = buf;
      while (*s == ' '||*s == '\t') s++;
      t = strchr(s, '{');
      if (t != NULL) *t = ' ';

/*
      if (ilog)
	{
	  FILE *dfile = fopen(ifilename,"a");
	  fprintf(dfile, "%s(%d): command = \"%s\"\n", __FILE__, __LINE__, s); 
	  fclose(dfile);
	}
*/
				/* parse the command */
      cmd = s;
      s += strcspn(s, " \t");
      if (*s) *s++ = 0;
      while (*s == ' ') s++;

      if (strcmp(cmd, "dectalk_capitalize") == 0)
	capitalize = atoi(s);

      else if (strcmp(cmd, "l") == 0)
	dectalk_letter(s);

      else if (strcmp(cmd, "dectalk_reset") == 0)
	dectalk_reset();

      else if (strcmp(cmd, "dectalk_say") == 0)
	dectalk_say(s);

      else if (strcmp(cmd, "dectalk_set_character_scale") == 0)
	dectalk_set_character_scale(atof(s));

      else if (strcmp(cmd, "dectalk_set_punctuations") == 0)
	dectalk_set_punctuations(s);

      else if (strcmp(cmd, "dectalk_set_speech_rate") == 0)
	dectalk_set_speech_rate(atoi(s));

      else if (strcmp(cmd, "dectalk_speak") == 0)
	dectalk_speak(s);

      else if (strcmp(cmd, "dectalk_split_caps") == 0)
	split_caps = atoi(s);

      else if (strcmp(cmd, "s") == 0)
	dectalk_stop();

      else if (strcmp(cmd, "t") == 0)
	dectalk_tone(s);

      else if (strcmp(cmd, "q") == 0)
	queue_speech(s);

      else if (strcmp(cmd, "exit") == 0)
	exit(0);
				/* old style commands (Emacspeak 5.0
                                   and earlier) */
      else if (strcmp(cmd, "dectalk_letter") == 0)
	dectalk_letter(s);
     
      else if (strcmp(cmd, "dectalk_stop") == 0)
	dectalk_stop();
     
      else if (strcmp(cmd, "dectalk_tone") == 0)
	dectalk_tone(s);
     
      else if (strcmp(cmd, "queue_speech") == 0)
	queue_speech(s);

      else 
	{
	  if (ilog)
	    {
	      FILE *dfile = fopen(ifilename,"a");
	      fprintf(dfile, "d-exp: unrecognized command \"%s\"\n", s); 
	      fclose(dfile);
	    }
	  fprintf(stderr, "d-exp: unrecognized command \"%s\"\n", s);
	  exit(1);
	}
    }
  return 0;
}

#ifdef NEVER
/* spare parts warehouse */
void
dectalk_pause()
{
  writeout("[:pause ]");
  talking=0;
}

void
dectalk_resume()
{
  writeout("[:resume ]");
  speech_task();
}

void dectalk_space_special_chars(int flag) {space_special_chars = flag;}

void
dectalk_synchronize()
{
  queue_speech("[:sync]");
}
#endif /* NEVER */


/* replacement for defective libraries */

#ifndef HAVE_STRSTR
	/* returns a pointer to the beginning of the first occurrence
	   of the substring needle in the string haystack. */
char *strstr(const char *haystack, const char *needle)
{				/* slow but simple implementation */
  char c;
  int hlen=strlen(haystack), nlen = strlen(needle), i, j;

  for (j = 0; j <= hlen-nlen; j++)
    {
      for (i = 0; (c = needle[i]); i++)
	if (haystack[j+i] != needle[i])
	  break;
      if (c == 0)
	return (char *)haystack + j;
    }
  return NULL;
}
#endif
